{*********************************************************}
{                                                         }
{     TParser 10.1 for Borland Delphi                     }
{                                                         }
{       A component for parsing and evaluating            }
{       mathematical expessions specified at runtime     }
{                                                        }
{         Renate Schaaf (schaaf@math.usu.edu), 1993       }
{         Alin Flaider (aflaidar@datalog.ro), 1996        }
{         Stefan Hoffmeister                              }
{              Stefan.Hoffmeister@Uni-Passau.de, 1997     }
{                                                         }
{                                                         }
{  See  PARSER10.TXT  for documentation                   }
{                                                         }
{*********************************************************}
unit Parser10;

{$IFDEF Win32}
  {$H+,S-} { long strings, no stack-checking}
{$ENDIF}

{.$DEFINE DEBUG} { by default make it lean and efficient }
{$IFNDEF DEBUG}
  {$D-} {$L-} {$Q-} {$R-} {$S-}
{$ENDIF}

{$I+} { I/O checking ON }

interface

uses
  SysUtils,
  Classes;

type
  { a couple of unfortunately necessary global declarations }
  ParserFloat = double;  { please do NOT use "real", only single, double, extended}
  PParserFloat = ^ParserFloat;

  TToken=( variab, constant,
           minus,
           sum, diff, prod, divis, modulo, IntDiv,
           integerpower, realpower,
           square, third, fourth,
           FuncOneVar, FuncTwoVar);

  POperation = ^TOperation;
  { functions that are added to the engine MUST have this declaration }
  { make sure that the procedure is declared far !!! }
  TMathProcedure = procedure(AnOperation: POperation);
  TOperation = record
                 { MUST use pointers (!), because argument and destination are linked... }
                 Arg1, Arg2 : PParserFloat;
                 Dest : PParserFloat;

                 NextOperation : POperation;

                 Operation: TMathProcedure;
                 Token : TToken;
               end;

  EMathParserError = class(Exception); { create a new exception class and... }

  { ... some descendants }
  ESyntaxError = class(EMathParserError);
  EExpressionHasBlanks = class(EMathParserError);
  EExpressionTooComplex = class(EMathParserError);
  ETooManyNestings = class(EMathParserError);
  EMissMatchingBracket = class(EMathParserError);
  EBadName = class(EMathParserError);
  EParserInternalError = class(EMathParserError); { hopefully we will never see this one }


  { we COULD use Forms and the TExceptionEvent therein,
    but that would give us all the VCL overhead.
    Consequentially we just redeclare an appropriate event }
  TParserExceptionEvent = procedure (Sender: TObject; E: Exception) of object;



  TCustomParser2 = class(TComponent)
  private
    { some pre-allocated space for variables }
    FA,
    FB,
    FC,
    FD,
    FE,
    FM,
    FN,
    FO,
    FP,
    FQ,
    FR,
    FS,
    FU,
    FV,
    FW,
    FZ,
    FX,
    FY,
    FT: ParserFloat;
  private
    FExpression : string;
    FPascalNumberformat: boolean;
    FParserError : boolean;
    FVariables: TStringList;

    FStartOperationList: POperation;

    FOnParserError : TParserExceptionEvent;

    function CalcValue: extended;
    procedure SetExpression(const AnExpression: string);
    procedure SetVar(const VarName: string; const Value: extended);
  protected
    { lists of available functions, see .Create for example use }
    FunctionOne : TStringList;     { functions with ONE argument, e.g. exp() }
    FunctionTwo : TStringList;     { functions with TWO arguments, e.g. max(,) }

    { predefined variables - could be left out }
    property A: ParserFloat read FA write FA;
    property B: ParserFloat read FB write FB;
    property C: ParserFloat read FC write FC;
    property D: ParserFloat read FD write FD;
    property E: ParserFloat read FE write FE;
    property M: ParserFloat read FM write FM;
    property N: ParserFloat read FN write FN;
    property O: ParserFloat read FO write FO;
    property P: ParserFloat read FP write FP;
    property Q: ParserFloat read FQ write FQ;
    property R: ParserFloat read FR write FR;
    property S: ParserFloat read FS write FS;
    property T: ParserFloat read FT write FT;
    property U: ParserFloat read FU write FU;
    property V: ParserFloat read FV write FV;
    property W: ParserFloat read FW write FW;
    property X: ParserFloat read FX write FX;
    property Y: ParserFloat read FY write FY;
    property Z: ParserFloat read FZ write FZ;
    
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;

    function ParseExpression(const AnExpression: string): boolean;
    procedure FreeExpression;

    { The PParserFloat returned points to the place in memory where the
      variable actually sits; to speed up assignment you can DIRECTLY
      assign data to the memory area. }
    function SetVariable(VarName: string; const Value: extended): PParserFloat;
    function GetVariable(const VarName: string): extended;

    procedure AddFunctionOneParam(const AFunctionName: string; const Func: TMathProcedure);
    procedure AddFunctionTwoParam(const AFunctionName: string; const Func: TMathProcedure);

    procedure ClearVariables;
    procedure ClearVariable(const AVarName: string);
    function  VariableExists(const AVarName: string): boolean;

    procedure ClearFunctions;
    procedure ClearFunction(const AFunctionName: string);

    property ParserError: boolean read FParserError;
    property LinkedOperationList: POperation read FStartOperationList;

    property Variable[const VarName: string]: extended read GetVariable write SetVar;
  published
    property Value: extended read CalcValue stored false;

    { setting Expression automatically parses it
      Warning: exceptions MAY be raised, if OnParserError is NOT assigned,
               otherwise the event will be triggered in case of an error }
    property Expression: string read FExpression write SetExpression;
    property PascalNumberformat: boolean read FPascalNumberformat write FPascalNumberformat default true;
    property OnParserError: TParserExceptionEvent read FOnParserError write FOnParserError;
  end;




  TParser = class(TCustomParser2)
  public
    { overrides to add the properties below as variables
      and adds all the functions }
    constructor Create(AOwner: TComponent); override;

    { returns the string with the blanks inside removed }
    class function RemoveBlanks(const s: string): string;
  published
    { predefined variables - could be left out }
    property A;
    property B;
    property C;
    property D;
    property E;
    property M;
    property N;
    property O;
    property P;
    property Q;
    property R;
    property S;
    property T;
    property U;
    property V;
    property W;
    property X;
    property Y;
    property Z;
 end;



procedure Register;

implementation

{$DEFINE UseMath}
{ Note: if you do not have the MATH unit simply remove the conditional define
        the component will continue to work, just a bit slower }

uses
{$IFDEF UseMath}
  Math,Main,
{$ENDIF}
  P10Build;

procedure Register;
begin
  RegisterComponents('Parsers', [TParser]);
end;
(*
{$IFDEF VER80}
  {$R *.D16}
{$ELSE}
  {$IFDEF VER90}
    {$R *.D32}
  {$ENDIF}
{$ENDIF}
*)

{****************************************************************}
{                                                                }
{   Following are "built-in" calculation procedures              }
{                                                                }
{****************************************************************}
{
Naming convention for functions:

  Name of built-in function, prepended with an underscore.
  Example:

    ln --> _ln

Passed arguments / results:

  If the function takes any arguments - i.e. if it has been added to
  either the FunctionOne or the FunctionTwo list:

  - First  argument --> arg1^
  - Second argument --> arg2^

  The result of the operation must ALWAYS be put into

     dest^


 Note: These are POINTERS to floats.
}



{****************************************************************}
{                                                                }
{   These are mandatory procedures - never remove them           }
{                                                                }
{****************************************************************}

{ do nothing - this only happens if the "term" is just a number
  or a variable; otherwise this procedure will never be called }
procedure _nothing(AnOp: POperation); far;
begin
end;

procedure _Add(AnOp: POperation); far;
begin
  with AnOp^ do
    dest^ := arg1^ + arg2^;
end;

procedure _Subtract(AnOp: POperation); far;
begin
  with AnOp^ do
    dest^ := arg1^ - arg2^;
end;

procedure _Multiply(AnOp: POperation); far;
begin
  with AnOp^ do
    dest^ := arg1^ * arg2^;
end;

procedure _RealDivide(AnOp: POperation); far;
begin
  with AnOp^ do
    dest^ := arg1^ / arg2^;
end;

procedure _Modulo(AnOp: POperation); far;
begin
  with AnOp^ do
    dest^ := trunc(arg1^) mod trunc(arg2^);
end;

procedure _IntDiv(AnOp: POperation); far;
begin
  with AnOp^ do
    dest^ := trunc(arg1^) div trunc(arg2^);
end;

procedure _Negate(AnOp: POperation); far;
begin
  with AnOp^ do
    dest^ := -arg1^;
end;

procedure _IntPower(AnOp: POperation); far;
{$IFNDEF UseMath}
var
  n, i: longint;
{$ENDIF}
begin

{$IFNDEF UseMath}
  with AnOp^ do
  begin
    n := trunc(abs(arg2^))-1;

    case n of
      -1: dest^ := 1;
       0: dest^ := arg1^;
    else
      dest^ := arg1^;
      for i := 1 to n do
        dest^ := dest^ * arg1^;
    end;

    if arg2^ < 0 then
      dest^ := 1 / dest^;

  end;
{$ELSE}
  with AnOp^ do
    dest^ := IntPower(arg1^, trunc(arg2^));
{$ENDIF}
end;

procedure _square(AnOp: POperation); far;
begin
  with AnOp^ do
    dest^ := sqr(arg1^);
end;

procedure _third(AnOp: POperation); far;
begin
  with AnOp^ do
    dest^ := arg1^ * arg1^ * arg1^;
end;

procedure _forth(AnOp: POperation); far;
begin
  with AnOp^ do
    dest^ := sqr(sqr(arg1^));
end;

procedure _power(AnOp: POperation); far;
begin
  with AnOp^ do
  begin
{$IFNDEF UseMath}
    if arg1^ = 0 then
      dest^ := 0
    else
      dest^ := exp(arg2^*ln(arg1^));
{$ELSE}
    dest^ := Power(arg1^, arg2^);
{$ENDIF}
  end;
end;


{****************************************************************}
{                                                                }
{   These are OPTIONAL procedures - you may remove them, though  }
{   it is preferable to not register them for use                }
{                                                                }
{****************************************************************}
procedure _arcsinh(AnOp: POperation); far;
begin
  with AnOp^ do
       dest^:=Ln(arg1^+sqrt(1+power(arg1^,2)));
end;

procedure _arccosh(AnOp: POperation); far;
begin
  with AnOp^ do
       dest^:=Ln(arg1^+sqrt(power(arg1^,2)-1));
end;

procedure _arctanh(AnOp: POperation); far;
begin
  with AnOp^ do
       dest^:=Ln((1+arg1^)/(1-arg1^))/2;
end;

procedure _arcsech(AnOp: POperation); far;
begin
  with AnOp^ do
       dest^:=Ln((1/arg1^)+sqrt(power((1/arg1^),2)-1));
end;

procedure _arccosech(AnOp: POperation); far;
begin
  with AnOp^ do
       dest^:=Ln((1/arg1^)+sqrt(1+power((1/arg1^),2)));
end;

procedure _arccoth(AnOp: POperation); far;
begin
  with AnOp^ do
       dest^:=Ln((1+(1/arg1^))/(1-(1/arg1^)))/2;
end;

procedure _arcsin(AnOp: POperation); far;
begin
  with AnOp^ do
       dest^:=arcsin(arg1^);
end;

procedure _degarcsin(AnOp: POperation); far;
begin
  with AnOp^ do
       dest^:=arcsin(DegToRad(arg1^));
end;

procedure _arccos(AnOp: POperation); far;
begin
  with AnOp^ do
       dest^:=arccos(arg1^);
end;

procedure _arcsec(AnOp: POperation); far;
begin
  with AnOp^ do
    dest^ := arccos(1/arg1^);
end;

procedure _degarcsec(AnOp: POperation); far;
begin
  with AnOp^ do
    dest^ := RadToDeg(arccos(1/arg1^));
end;

procedure _arccosec(AnOp: POperation); far;
begin
  with AnOp^ do
    dest^ := arcsin(1/arg1^);
end;

procedure _degarccosec(AnOp: POperation); far;
begin
  with AnOp^ do
    dest^ := RadToDeg(arcsin(1/arg1^));
end;

procedure _degarccos(AnOp: POperation); far;
begin
  with AnOp^ do
       dest^:=arccos(DegToRad(arg1^));
end;

procedure _sin(AnOp: POperation); far;
begin
  with AnOp^ do
         dest^ := sin(arg1^);
end;

procedure _sinsq(AnOp: POperation); far;
begin
  with AnOp^ do
         dest^ := power(sin(arg1^),2);
end;

procedure _sinsqrad(AnOp: POperation); far;
begin
  with AnOp^ do
         dest^ := power(sin(DegToRad(arg1^)),2);
end;

procedure _sincube(AnOp: POperation); far;
begin
  with AnOp^ do
         dest^ := power(sin(arg1^),3);
end;

procedure _sincuberad(AnOp: POperation); far;
begin
  with AnOp^ do
         dest^ := power(sin(DegToRad(arg1^)),3);
end;

procedure _cos(AnOp: POperation); far;
begin
  with AnOp^ do
    dest^ := cos(arg1^);
end;

procedure _cossq(AnOp: POperation); far;
begin
  with AnOp^ do
    dest^ := sqr(cos(arg1^));
end;

procedure _cossqrad(AnOp: POperation); far;
begin
  with AnOp^ do
    dest^ := sqr(cos(DegToRad(arg1^)));
end;

procedure _coscube(AnOp: POperation); far;
begin
  with AnOp^ do
    dest^ := power(cos(arg1^),3);
end;

procedure _coscuberad(AnOp: POperation); far;
begin
  with AnOp^ do
    dest^ := power(cos(DegToRad(arg1^)),3);
end;

procedure _arctan(AnOp: POperation); far;
begin
  with AnOp^ do
    dest^ := arctan(arg1^);
end;

procedure _arccot(AnOp: POperation); far;
begin
  with AnOp^ do
    dest^ := arctan(1/arg1^);
end;

procedure _degarccot(AnOp: POperation); far;
begin
  with AnOp^ do
    dest^ := RadToDeg(arctan(1/arg1^));
end;

procedure _cosechsq(AnOp: POperation); far;
begin
  with AnOp^ do
    dest^ := power(2/(exp(arg1^)-exp(-arg1^)),2);
end;

procedure _cosechcube(AnOp: POperation); far;
begin
  with AnOp^ do
    dest^ := power(2/(exp(arg1^)-exp(-arg1^)),3);
end;

procedure _cosech(AnOp: POperation); far;
begin
  with AnOp^ do
    dest^ := 2/(exp(arg1^)-exp(-arg1^));
end;

procedure _sechsq(AnOp: POperation); far;
begin
  with AnOp^ do
    dest^ := power(2/(exp(arg1^)+exp(-arg1^)),2);
end;

procedure _sechcube(AnOp: POperation); far;
begin
  with AnOp^ do
    dest^ := power(2/(exp(arg1^)+exp(-arg1^)),3);
end;

procedure _sech(AnOp: POperation); far;
begin
  with AnOp^ do
    dest^ := 2/(exp(arg1^)+exp(-arg1^));
end;

procedure _tanhsq(AnOp: POperation); far;
begin
  with AnOp^ do
    dest^ := power((exp(arg1^)-exp(arg1^))/(exp(arg1^)+exp(-arg1^)),2);
end;

procedure _tanhcube(AnOp: POperation); far;
begin
  with AnOp^ do
    dest^ :=power((exp(arg1^)-exp(arg1^))/(exp(arg1^)+exp(-arg1^)),3);
end;

procedure _tanh(AnOp: POperation); far;
begin
  with AnOp^ do
    dest^ := (exp(arg1^)-exp(arg1^))/(exp(arg1^)+exp(-arg1^));
end;

procedure _cothsq(AnOp: POperation); far;
begin
  with AnOp^ do
    dest^ := power((exp(arg1^)+exp(arg1^))/(exp(arg1^)-exp(-arg1^)),2);
end;

procedure _cothcube(AnOp: POperation); far;
begin
  with AnOp^ do
    dest^ :=power((exp(arg1^)+exp(arg1^))/(exp(arg1^)-exp(-arg1^)),3);
end;

procedure _coth(AnOp: POperation); far;
begin
  with AnOp^ do
    dest^ := (exp(arg1^)+exp(arg1^))/(exp(arg1^)-exp(-arg1^));
end;

procedure _degarctan(AnOp: POperation); far;
begin
  with AnOp^ do
    dest^ := RadToDeg(arctan(arg1^));
end;

procedure _arg(AnOp: POperation); far;
begin
  with AnOp^ do
    if arg1^ < 0 then
      dest^ := arctan(arg2^/arg1^)+Pi
    else
      if arg1^>0 then
        dest^ := arctan(arg2^/arg1^)
      else
        if arg2^ > 0 then
          dest^ := 0.5 * Pi
        else
          dest^ := -0.5 * Pi;
end;

procedure _sinh(AnOp: POperation); far;
begin
  with AnOp^ do
    dest^ := (exp(arg1^)-exp(-arg1^))*0.5;
end;

procedure _sinhsq(AnOp: POperation); far;
begin
  with AnOp^ do
    dest^ := power((exp(arg1^)-exp(-arg1^))*0.5,2);
end;

procedure _sinhcube(AnOp: POperation); far;
begin
  with AnOp^ do
    dest^ := power((exp(arg1^)-exp(-arg1^))*0.5,3);
end;

procedure _cosh(AnOp: POperation); far;
begin
  with AnOp^ do
    dest^ := (exp(arg1^)+exp(-arg1^))*0.5;
end;

procedure _coshsq(AnOp: POperation); far;
begin
  with AnOp^ do
    dest^ := power(exp(arg1^)+exp(-arg1^)*0.5,2);
end;
procedure _coshcube(AnOp: POperation); far;
begin
  with AnOp^ do
    dest^ := power(exp(arg1^)+exp(-arg1^)*0.5,2);
end;

procedure _cot(AnOp: POperation); far;
begin
  with AnOp^ do
      if Abs(arg1^)>0.000000001 then dest^ := cotan(arg1^) else dest^:=100000;
end;

procedure _cotrad(AnOp: POperation); far;
begin
  with AnOp^ do
    dest^ := cotan(DegToRad(arg1^));
end;

procedure _cotsq(AnOp: POperation); far;
begin
  with AnOp^ do
    dest^ := sqr(cotan(arg1^));
end;

procedure _cotsqrad(AnOp: POperation); far;
begin
  with AnOp^ do
    dest^ := sqr(cotan(DegToRad(arg1^)));
end;

procedure _cotcube(AnOp: POperation); far;
begin
  with AnOp^ do
    dest^ := power(cotan(arg1^),3);
end;

procedure _cotcuberad(AnOp: POperation); far;
begin
  with AnOp^ do
    dest^ := power(cotan(DegToRad(arg1^)),3);
end;


procedure _sec(AnOp: POperation); far;
begin
  with AnOp^ do
    dest^ := 1/cos(arg1^);
end;

procedure _cosec(AnOp: POperation); far;
begin
  with AnOp^ do
   if Abs(arg1^)>0.000000001 then dest^ := 1/sin(arg1^) else dest^:=1000000;
end;

procedure _tan(AnOp: POperation); far;
begin
  with AnOp^ do
    dest^ := tan(arg1^);
end;

procedure _tanrad(AnOp: POperation); far;
begin
  with AnOp^ do
    dest^ := tan(DegToRad(arg1^));
end;

procedure _tansq(AnOp: POperation); far;
begin
  with AnOp^ do
    dest^ := power(tan(arg1^),2);
end;

procedure _tansqrad(AnOp: POperation); far;
begin
  with AnOp^ do
    dest^ := power(tan(DegToRad(arg1^)),2);
end;

procedure _tancube(AnOp: POperation); far;
begin
  with AnOp^ do
    dest^ := power(tan(arg1^),3);
end;

procedure _tancuberad(AnOp: POperation); far;
begin
  with AnOp^ do
    dest^ := power(tan(DegToRad(arg1^)),3);
end;

procedure _sinrad(AnOp: POperation); far;
begin
  with AnOp^ do
    dest^ := sin(DegToRad(arg1^));
end;

procedure _cosrad(AnOp: POperation); far;
begin
  with AnOp^ do
    dest^ := cos(DegToRad(arg1^));
end;

procedure _cosecsq(AnOp: POperation); far;
begin
  with AnOp^ do
    dest^ := sqr(1/sin(arg1^));
end;

procedure _cosecsqrad(AnOp: POperation); far;
begin
  with AnOp^ do
    dest^ := sqr(1/sin(DegToRad(arg1^)));
end;

procedure _coseccube(AnOp: POperation); far;
begin
  with AnOp^ do
    dest^ := power(1/sin(arg1^),3);
end;

procedure _coseccuberad(AnOp: POperation); far;
begin
  with AnOp^ do
    dest^ := power(1/sin(DegToRad(arg1^)),3);
end;

procedure _cosecrad(AnOp: POperation); far;
begin
  with AnOp^ do
    dest^ := 1/sin(DegToRad(arg1^));
end;

procedure _secsq(AnOp: POperation); far;
begin
  with AnOp^ do
    dest^ := sqr(1/cos(arg1^));
end;

procedure _secsqrad(AnOp: POperation); far;
begin
  with AnOp^ do
    dest^ := sqr(1/cos(DegToRad(arg1^)));
end;

procedure _seccube(AnOp: POperation); far;
begin
  with AnOp^ do
    dest^ := power(1/cos(arg1^),3);
end;

procedure _seccuberad(AnOp: POperation); far;
begin
  with AnOp^ do
    dest^ := power(1/cos(DegToRad(arg1^)),3);
end;

procedure _secrad(AnOp: POperation); far;
begin
  with AnOp^ do
    dest^ := 1/cos(DegToRad(arg1^));
end;

procedure _exp(AnOp: POperation); far;
begin
  with AnOp^ do
    dest^ := exp(arg1^);
end;

procedure _ln(AnOp: POperation); far;
begin
  with AnOp^ do
    If arg1^>0 then dest^ := ln(arg1^) else dest^:=-200000;
end;

procedure _log10(AnOp: POperation); far;
const
  _1_ln10 =  0.4342944819033;
begin
  with AnOp^ do
{$IFDEF UseMath}
    If arg1^>0 then dest^ := log10(arg1^) else dest^:=-200000;
{$ELSE}
    If arg1^>0 then dest^ := ln(arg1^) * _1_ln10 else dest^:=-200000;
{$ENDIF}
end;

procedure _log2(AnOp: POperation); far;
const
  _1_ln2 = 1.4426950409;
begin
  with AnOp^ do
{$IFDEF UseMath}
    If arg1^>0 then dest^ := log2(arg1^)  else dest^:=-200000;
{$ELSE}
    If arg1^>0 then dest^ := ln(arg1^) * _1_ln2  else dest^:=-200000;
{$ENDIF}
end;

procedure _logN(AnOp: POperation); far;
begin
  with AnOp^ do
{$IFDEF UseMath}
    If arg1^>0 then dest^ := logN(arg1^, arg2^)  else dest^:=2000;
{$ELSE}
    If arg1^>0 then dest^ := ln(arg1^) / ln(arg2^)  else dest^:=2000;
{$ENDIF}
end;

procedure _sqrt(AnOp: POperation); far;
begin
  with AnOp^ do
    dest^ := sqrt(arg1^);
end;

procedure _blanc(AnOp: POperation);far;
Var
   temp,result:extended;
   count:integer;
begin
 with AnOp^ do
  Begin
  result:=0;
   for count := 0 to 49 do
   begin
     arg1^:=arg1^-int(arg1^-(Ord(arg1^<0)));
     if (arg1^)<0.5 then temp:=(arg1^) else temp:=(1-arg1^);
     result:=result+ temp/(power(2,count));
     arg1^:=2*arg1^;
   end;
   dest^:=result;
  end;
end;

procedure _abs(AnOp: POperation); far;
begin
  with AnOp^ do
    dest^ := abs(arg1^);
end;

procedure _min(AnOp: POperation); far;
begin
  with AnOp^ do
    if arg1^ < arg2^ then
      dest^ := arg1^
    else
      dest^ := arg2^;
end;

procedure _max(AnOp: POperation); far;
begin
  with AnOp^ do
    if arg1^ < arg2^ then
      dest^ := arg2^
    else
      dest^ := arg1^;
end;

procedure _heaviside(AnOp: POperation); far;
begin
  with AnOp^ do
    if arg1^ < 0 then
      dest^ := 0
    else
      dest^ := 1;
end;

procedure _sign(AnOp: POperation); far;
begin
  with AnOp^ do
    if arg1^ < 0 then
      dest^ := -1
    else
      if arg1^ > 0 then
        dest^ := 1.0
      else
        dest^ := 0.0;
end;

procedure _zero(AnOp: POperation); far;
begin
  with AnOp^ do
    if arg1^ = 0.0 then
      dest^ := 0.0
    else
      dest^ := 1.0;
end;

procedure _trunc(AnOp: POperation); far;
begin
  with AnOp^ do
    dest^ := int(arg1^)
end;

procedure _ceil(AnOp: POperation); far;
begin
  with AnOp^ do
    if frac(arg1^) > 0 then
      dest^ := int(arg1^ + 1)
    else
      dest^ := int(arg1^);
end;

procedure _floor(AnOp: POperation); far;
begin
  with AnOp^ do
    if frac(arg1^) < 0 then
      dest^ := int(arg1^ - 1)
    else
      dest^ := int(arg1^);
end;

procedure _rnd(AnOp: POperation); far;
begin
  with AnOp^ do
    dest^ := Random * int(arg1^);
end;

procedure _random(AnOp: POperation); far;
begin
  with AnOp^ do
    dest^ := Random;
end;

procedure _radius(AnOp: POperation); far;
begin
  with AnOp^ do
    dest^ := sqrt(sqr(arg1^)+sqr(arg2^));
end;

procedure _phase(AnOp: POperation); far;
var
  a: ParserFloat;
begin
  with AnOp^ do
  begin
    a := arg1^ / (2/pi);
    dest^ := (2*pi) * (a-round(a));
  end;
end;

procedure _function1(AnOp: POperation); far;
Begin
  with AnOp^ do
     Begin
          Mainform.FParser1.X:=arg1^;
          dest^:=Mainform.FParser1.Value;
     end;
end;

procedure _function2(AnOp: POperation); far;
Begin
  with AnOp^ do
     Begin
          Mainform.FParser2.X:=arg1^;
          dest^:=Mainform.FParser2.Value;
     end;
end;

procedure _function3(AnOp: POperation); far;
Begin
  with AnOp^ do
     Begin
          Mainform.FParser3.X:=arg1^;
          dest^:=Mainform.FParser3.Value;
     end;
end;

procedure _function4(AnOp: POperation); far;
Begin
  with AnOp^ do
     Begin
          Mainform.FParser4.X:=arg1^;
          dest^:=Mainform.FParser4.Value;
     end;
end;

procedure _function5(AnOp: POperation); far;
Begin
  with AnOp^ do
     Begin
          Mainform.FParser5.X:=arg1^;
          dest^:=Mainform.FParser5.Value;
     end;
end;

procedure _function6(AnOp: POperation); far;
Begin
  with AnOp^ do
     Begin
          Mainform.FParser6.X:=arg1^;
          dest^:=Mainform.FParser6.Value;
     end;
end;

procedure _function7(AnOp: POperation); far;
Begin
  with AnOp^ do
     Begin
          Mainform.FParser7.X:=arg1^;
          dest^:=Mainform.FParser7.Value;
     end;
end;

procedure _diff(AnOp: POperation); far;
Begin
  Mainform.derivative:=True;
  AnOp^.dest^:=AnOp^.arg1^;
end;

procedure _integ(AnOp: POperation); far;
Begin
  Mainform.integral:=True;
  AnOp^.dest^:=AnOp^.arg1^;
end;

procedure _inv(AnOp: POperation); far;
Begin
  Mainform.inverse:=True;
  AnOp^.dest^:=AnOp^.arg1^;
end;

{****************************************************************}
{                                                                }
{   TCustomParser2                                                }
{                                                                }
{    A base class which does not publish the variable properties }
{    and adds no functions by default                            }
{                                                                }
{****************************************************************}
function TCustomParser2.ParseExpression(const AnExpression: string):boolean;
var
  OperationLoop: POperation;
begin
  FreeExpression;
  FExpression := AnExpression;

  if AnExpression <> '' then
  begin
    Result := false;

    try
      ParseFunction( AnExpression,

                     FVariables,

                     FunctionOne,
                     FunctionTwo,

                     FPascalNumberformat,

                     FStartOperationList,
                     Result);

      FParserError := Result;

    except
      on E: Exception do
      begin
        FParserError := true;

        if Assigned(FOnParserError) then
        begin
          FOnParserError(Self, E);
          exit;
        end
        else
          raise;
      end;
    end;

    Result := not Result;

    OperationLoop := FStartOperationList;
    while OperationLoop <> nil do
    begin
      with OperationLoop^ do
      begin
        case Token of

          variab,
          constant:      Operation := _nothing;

          minus:         Operation := _negate;

          sum:           Operation := _add;
          diff:          Operation := _subtract;
          prod:          Operation := _multiply;
          divis:         Operation := _RealDivide;
          
          modulo:        Operation := _Modulo;
          intdiv:        Operation := _IntDiv;

          integerpower:  Operation := _IntPower;
          realpower:     Operation := _Power;

          square:        Operation := _square;
          third:         Operation := _third;
          fourth:        Operation := _forth;

          FuncOneVar, FuncTwoVar:    { job has been done in build already !};
        end; {case}

        OperationLoop := NextOperation;
      end; {with OperationLoop^}

    end; {while OperationLoop<>nil}
  end;
end;

constructor TCustomParser2.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);

  FPascalNumberformat := true;

  FVariables := TStringList.Create;
  with FVariables do
  begin
    Sorted := true;
    Duplicates := dupIgnore;
  end;

  FunctionOne := TStringList.Create;
  with FunctionOne do
  begin
    Sorted := true;
    Duplicates := dupError;
  end;

  FunctionTwo := TStringList.Create;
  with FunctionTwo do
  begin
    Sorted := true;
    Duplicates := dupError;
  end;

end;

destructor TCustomParser2.Destroy;
begin
  FreeExpression;

  ClearVariables;
  FVariables.Free;

  FunctionOne.Free;
  FunctionTwo.Free;

  inherited Destroy;
end;




procedure TCustomParser2.SetVar(const VarName: string; const Value: extended);
begin
  SetVariable(VarName, Value);
end;

function TCustomParser2.SetVariable(VarName: string; const Value: extended): PParserFloat;
var
  i: integer;
begin
  { always convert to uppercase }
  VarName := UpperCase(VarName);

  with FVariables do
    if Find(VarName, i) then
    begin
      Result := PParserFloat(Objects[i]);
      Result^ := Value;
    end
    else
    begin
      if Length(Varname) = 1 then
        case VarName[1] of
          'A': Result := @FA;
          'B': Result := @FB;
          'C': Result := @FC;
          'D': Result := @FD;
          'E': Result := @FE;
          'M': Result := @FM;
          'N': Result := @FN;
          'O': Result := @FO;
          'P': Result := @FP;
          'Q': Result := @FQ;
          'R': Result := @FR;
          'S': Result := @FS;
          'T': Result := @FT;
          'U': Result := @FU;
          'V': Result := @FV;
          'W': Result := @FW;
          'X': Result := @FX;
          'Y': Result := @FY;
          'Z': Result := @FZ;
        else { case }
          { is the variable name a valid identifier? }
          if not IsValidIdent(VarName) then
            raise EBadName.Create(VarName);

          { unravelled loop for improved (string!) performance! }

          { check whether the variable contains any of the operators (DIV and MOD)
            this would confuse the parser... }
          if pos('+', VarName) <> 0 then
              raise EBadName.Create(VarName);

          if pos('-', VarName) <> 0 then
              raise EBadName.Create(VarName);

          if pos('*', VarName) <> 0 then
              raise EBadName.Create(VarName);

          if pos('/', VarName) <> 0 then
              raise EBadName.Create(VarName);

          if pos('^', VarName) <> 0 then
              raise EBadName.Create(VarName);

          if pos('DIV', VarName) <> 0 then
              raise EBadName.Create(VarName);

          if pos('MOD', VarName) <> 0 then
              raise EBadName.Create(VarName);

          new(Result);
        end { case }
      else
      begin
        { is the variable name a valid identifier? }
        if not IsValidIdent(VarName) then
          raise EBadName.Create(VarName);

        new(Result);
      end;

      Result^ := Value;

      AddObject(VarName, TObject(Result));
    end
end;

function TCustomParser2.GetVariable(const VarName: string): extended;
var
  i: integer;
begin
  with FVariables do
    if Find(UpperCase(VarName), i) then
      Result := PParserFloat(Objects[i])^
    else
      Result := 0.0;
end;

procedure TCustomParser2.AddFunctionOneParam(const AFunctionName: string; const Func: TMathProcedure);
begin
  if IsValidIdent(AFunctionName) then
    FunctionOne.AddObject(UpperCase(AFunctionName), TObject(@Func))
  else
    raise EBadName.Create(AFunctionName);
end;

procedure TCustomParser2.AddFunctionTwoParam(const AFunctionName: string; const Func: TMathProcedure);
begin
  if IsValidIdent(AFunctionName) then
    FunctionTwo.AddObject(UpperCase(AFunctionName), TObject(@Func))
  else
    raise EBadName.Create(AFunctionName);
end;

procedure TCustomParser2.ClearVariables;
var
  i: integer;
  APPFloat: PParserFloat;
  AString: string; { disregard stack consumption }
begin
  with FVariables do
  begin
    i := Count;
    while i > 0 do
    begin
      dec(i);
      AString := Strings[i];
      if (Length(AString) <> 1) or
         (not (AString[1] in ['A'..'E','M'..'Z'])) then
      begin
        APPFloat := PParserFloat(Objects[i]);
        if APPFloat <> nil then
          dispose( APPFloat ); { dispose only user-defined variables }
      end;
    end;

    Clear;
  end;

  with FVariables do
  begin
    i := Count;
    while i > 0 do
    begin
      dec(i);
      AString := Strings[i];
      if (Length(AString) <> 1) or
         (not (AString[1] in ['A'..'E', 'M'..'Z'])) then
      begin
        APPFloat := PParserFloat(Objects[i]);
        if APPFloat <> nil then
          dispose( APPFloat ); { dispose only user-defined variables }
      end;
    end;

    Clear;
  end;

  SetExpression(''); { invalidate expression }
end;

procedure TCustomParser2.ClearVariable(const AVarName: string);
var
  index: integer;
begin
  with FVariables do
  begin
    if Find(AVarName, index) then
    begin
      if (Length(AVarName) <> 1) and
         (not (AVarName[1] in ['A'..'E', 'M'..'Z'])) then
        dispose( PParserFloat(Objects[index]) ); { dispose only user-defined variables }

      Delete(index);
    end;
  end;

  SetExpression(''); { invalidate expression }
end;

function TCustomParser2.VariableExists(const AVarName: string): boolean;
var
  index: integer;
begin
  Result := FVariables.Find(UpperCase(AVarName), index);
end;

procedure TCustomParser2.ClearFunctions;
begin
  FunctionOne.Clear;
  FunctionTwo.Clear;

  SetExpression(''); { invalidate expression }
end;

procedure TCustomParser2.ClearFunction(const AFunctionName: string);
var
  index: integer;
begin
  with FunctionOne do
  begin
    if Find(AFunctionName, index) then
    begin
      Delete(index);
      SetExpression(''); { invalidate expression }
      exit;
    end;
  end;

  with FunctionTwo do
  begin
    if Find(AFunctionName, index) then
    begin
      Delete(index);
      SetExpression(''); { invalidate expression }
    end;
  end;
end;


procedure TCustomParser2.FreeExpression;
var
  LastOP,
  NextOP: POperation;
begin
  LastOP := FStartOperationList;

  while LastOP <> nil do
  begin
    NextOP := LastOP^.NextOperation;

    while NextOP <> nil do
      with NextOP^ do
      begin
        if (Arg1 = lastop^.Arg1) or (Arg1 = lastop^.Arg2) or (Arg1 = lastop^.Dest) then
          Arg1 := nil;

        if (Arg2 = lastop^.Arg1) or (Arg2 = lastop^.Arg2) or (Arg2 = lastop^.Dest) then
          Arg2 := nil;

        if (Dest = lastop^.Arg1) or (Dest = lastop^.Arg2) or (Dest = lastop^.Dest) then
          Dest := nil;

        NextOP := NextOperation;
      end;

    with LastOP^, FVariables do
    begin
      if IndexOfObject( TObject(Arg1)) >= 0 then Arg1 := nil;
      if IndexOfObject( TObject(Arg2)) >= 0 then Arg2 := nil;
      if IndexOfObject( TObject(Dest)) >= 0 then Dest := nil;

      if (Dest <> nil) and (Dest <> Arg2) and (Dest <> Arg1) then
        dispose(Dest);

      if (Arg2 <> nil) and (Arg2 <> Arg1) then
        dispose(Arg2);

      if (Arg1 <> nil) then
        dispose(Arg1);
    end;

    NextOP := LastOP^.NextOperation;
    dispose(LastOP);
    LastOP := NextOP;
  end;

  FStartOperationList := nil;
end;

procedure TCustomParser2.SetExpression(const AnExpression: string);
begin
  ParseExpression(AnExpression); { this implies FExpression := AnExpression }
end;


function TCustomParser2.CalcValue: extended;
var
  LastOP: POperation;
begin
  if FStartOperationList <> nil then
  begin
    LastOP := FStartOperationList;

    while LastOP^.NextOperation <> nil do
    begin
      with LastOP^ do
      begin
        Operation(LastOP);
        LastOP := NextOperation;
      end;
    end;
    LastOP^.Operation(LastOP);

    Result := LastOP^.Dest^;
  end
  else
    Result := 0;
end;


{****************************************************************}
{                                                                }
{   TCustomParser2                                                }
{                                                                }
{****************************************************************}
constructor TParser.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);

  with FVariables do
  begin
    AddObject( 'A', TObject(@FA));
    AddObject( 'B', TObject(@FB));
    AddObject( 'C', TObject(@FC));
    AddObject( 'D', TObject(@FD));
    AddObject( 'E', TObject(@FE));
    AddObject( 'M', TObject(@FM));
    AddObject( 'N', TObject(@FN));
    AddObject( 'O', TObject(@FO));
    AddObject( 'P', TObject(@FP));
    AddObject( 'Q', TObject(@FQ));
    AddObject( 'R', TObject(@FR));
    AddObject( 'S', TObject(@FS));
    AddObject( 'T', TObject(@FT));
    AddObject( 'U', TObject(@FU));
    AddObject( 'V', TObject(@FV));
    AddObject( 'W', TObject(@FW));
    AddObject( 'X', TObject(@FX));
    AddObject( 'Y', TObject(@FY));
    AddObject( 'Z', TObject(@FZ));
  end;

  with FunctionOne do
  begin

    AddObject('ASINH', TObject(@_arcsinh));
    AddObject('ACOSH', TObject(@_arccosh));
    AddObject('ATANH', TObject(@_arctanh));
    AddObject('ASECH', TObject(@_arcsech));
    AddObject('ACOSECH', TObject(@_arccosech));
    AddObject('ACOTH', TObject(@_arccoth));
    AddObject('ASIN', TObject(@_arcsin));
    AddObject('DEGASIN', TObject(@_degarcsin));
    AddObject('ACOS', TObject(@_arccos));
    AddObject('DEGACOS', TObject(@_degarccos));
    AddObject('ATAN', TObject(@_arctan));
    AddObject('DEGATAN', TObject(@_degarctan));
    AddObject('ACOT', TObject(@_arccot));
    AddObject('DEGACOT', TObject(@_degarccot));
    AddObject('ASEC', TObject(@_arcsec));
    AddObject('DEGASEC', TObject(@_degarcsec));
    AddObject('ACOSEC', TObject(@_arccosec));
    AddObject('DEGACOSEC', TObject(@_degarccosec));

    AddObject('SINHSQ', TObject(@_sinhsq));
    AddObject('SINHCUBE', TObject(@_sinhcube));
    AddObject('SINH', TObject(@_sinh));
    AddObject('COSHSQ', TObject(@_coshsq));
    AddObject('COSHCUBE', TObject(@_coshcube));
    AddObject('COSH', TObject(@_cosh));
    AddObject('COSECHSQ', TObject(@_cosechsq));
    AddObject('COSECHCUBE', TObject(@_cosechcube));
    AddObject('COSECH', TObject(@_cosech));
    AddObject('SECHSQ', TObject(@_sechsq));
    AddObject('SECHCUBE', TObject(@_sechcube));
    AddObject('SECH', TObject(@_sech));
    AddObject('TANHSQ', TObject(@_tanhsq));
    AddObject('TANHCUBE', TObject(@_tanhcube));
    AddObject('TANH', TObject(@_tanh));
    AddObject('COTHSQ', TObject(@_cothsq));
    AddObject('COTHCUBE', TObject(@_cothcube));
    AddObject('COTH', TObject(@_coth));



    AddObject('TAN', TObject(@_tan));
    AddObject('TANRAD',TObject(@_tanrad));
    AddObject('SIN', TObject(@_sin));
    AddObject('SINRAD', TObject(@_sinrad));
    AddObject('COS', TObject(@_cos));
    AddObject('COSRAD', TObject(@_cosrad));

    AddObject('COSECSQ',TObject(@_cosecsq));
    AddObject('COSECSQRAD',TObject(@_cosecsqrad));
    AddObject('COSECCUBE',TObject(@_coseccube));
    AddObject('COSECCUBERAD',TObject(@_coseccuberad));
    AddObject('COSEC',TObject(@_cosec));
    AddObject('COSECRAD',TObject(@_cosecrad));
    AddObject('SECSQ', TObject(@_secsq));
    AddObject('SECSQRAD', TObject(@_secsqrad));
    AddObject('SECCUBE', TObject(@_seccube));
    AddObject('SECCUBERAD', TObject(@_seccuberad));
    AddObject('SEC', TObject(@_sec));
    AddObject('SECRAD', TObject(@_secrad));
    AddObject('COSSQ', TObject(@_cossq));
    AddObject('COSSQRAD', TObject(@_cossqrad));
    AddObject('COSCUBE', TObject(@_coscube));
    AddObject('COSCUBERAD', TObject(@_coscuberad));
    AddObject('SINSQ', TObject(@_sinsq));
    AddObject('SINSQRAD', TObject(@_sinsqrad));
    AddObject('SINCUBE', TObject(@_sincube));
    AddObject('SINCUBERAD', TObject(@_sincuberad));
    AddObject('TANSQ', TObject(@_tansq));
    AddObject('TANSQRAD', TObject(@_tansqrad));
    AddObject('TANCUBE', TObject(@_tancube));
    AddObject('TANCUBERAD', TObject(@_tancuberad));

    AddObject('COTSQ', TObject(@_cotsq));
    AddObject('COTSQRAD', TObject(@_cotsqrad));
    AddObject('COTCUBE', TObject(@_cotcube));
    AddObject('COTCUBERAD', TObject(@_cotcuberad));
    AddObject('COT', TObject(@_cot));
    AddObject('COTRAD', TObject(@_cotrad));

    AddObject('INTEG', TObject(@_integ));
    AddObject('DIFF', TObject(@_diff));
    AddObject('INV',TObject(@_inv));
    AddObject('LOG', TObject(@_log10));
    AddObject('LN', TObject(@_ln));
    AddObject('LOG2', TObject(@_log2));
    AddObject('SQRT', TObject(@_sqrt));
    AddObject('SQR',TObject(@_square));
    AddObject('INT', TObject(@_trunc));
    AddObject('EXP', TObject(@_exp));
    AddObject('ABS', TObject(@_abs));
    AddObject('BLANC', TObject(@_blanc));
    AddObject('ARG', TObject(@_arg));
    AddObject('TRUNC', TObject(@_trunc));
    AddObject('CEIL', TObject(@_ceil));
    AddObject('FLOOR', TObject(@_floor));
    AddObject('HEAV', TObject(@_heaviside));
    AddObject('SIGN', TObject(@_sign));
    AddObject('ZERO', TObject(@_zero));
    AddObject('PH', TObject(@_phase));
    AddObject('RND', TObject(@_rnd));
    AddObject('F',TObject(@_function1));
    AddObject('G',TObject(@_function2));
    AddObject('H',TObject(@_function3));
    AddObject('I',TObject(@_function4));
    AddObject('J',TObject(@_function5));
    AddObject('K',TObject(@_function6));
    AddObject('L',TObject(@_function7));
    AddObject('RANDOM', TObject(@_random));

  end;

  with FunctionTwo do
  begin

    AddObject('MAX', TObject(@_max));
    AddObject('MIN', TObject(@_min));


    AddObject('POWER', TObject(@_Power));
    AddObject('INTPOWER', TObject(@_IntPower));

    AddObject('LOGN', TObject(@_logN));

  end;
end;


class function TParser.RemoveBlanks(const s: string): string;
{deletes all blanks in s}
var
  i : integer;
begin
  Result := s;

  i := pos(' ', Result);
  while i > 0 do
  begin
    delete(Result, i, 1);
    i := pos(' ', Result);
  end;
end;

end.
